/*
モータの速度制御を行うモジュール
・PI制御により、左右のタイヤ回転速度（周速度）を指定した値に保つ。
・現在の速度値を提供する。
2014/07/14   by S. Ushimaru
*/

#include <stdio.h>
#include <linux/errno.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/io.h>
#include "hardware.h"
#include "encoder.h"
#include "motor.h"

static pthread_t tasks[1];
static float vel_l_ref, vel_r_ref;
static float vel_l_ref_org, vel_r_ref_org;
static float pid_controller_l(float, float);
static float pid_controller_r(float, float);

// PID 速度制御のゲイン
//左
static float Kp_l = 0.090;//0.0275;//0.0275; //1.2 ;  // 0.3
static float Ki_l = 0.0026;//0.0027;//0.0022; //0.085 ; //0.07 ; // 0.05 
static float Kd_l = 0.0 ; //
//右
static float Kp_r =  0.093;//0.0275;//0.0275; //1.0
static float Ki_r =  0.0025;//0.0025;//0.0022; // 0.05 0.1
static float Kd_r = 0.0 ; //

static float err_sum_l=0;
static float err_sum_r=0;
static float err_prev_l=0;
static float err_prev_r=0;
static float vel_l=0, vel_r=0;

static float t_coef = 2*3.1415*TIRE_R/ENC_PPS/T_ENCODER*1000/M_E_RATIO; //for STANDARD MIRS

static void *thread_code(void *t){

	int counter=0;
	static float u_l=0, u_r=0;
	struct encoder_data encoder_data;

	//printf("t_coef=%f\n",t_coef);
	//printf("enc_left \t enc_right \t  enc_left_sum \t enc_right_sum\n");

	while(1){
		counter++;

		encoder_get(&encoder_data);
		//printf("left=%5d  right=%5d left_sum=%5d right_sum=%5d\n",encoder_data.left,encoder_data.right, encoder_data.left_sum, encoder_data.right_sum);
		//printf("enc_left=%5d  enc_right=%5d \n",encoder_data.left,encoder_data.right);

		//t_coef=1;
		vel_l = encoder_data.left * t_coef * TIRE_DIST_RATIO;
		vel_r = encoder_data.right * t_coef;

		//printf("vel_l=%6.1f, vel_r=%6.1f   in vel_ctrl \n",vel_l, vel_r);

		if( vel_l_ref_org == 0 && vel_r_ref_org == 0){
			motor_drive(0, 0);
		}else{

			u_l = pid_controller_l(vel_l, vel_l_ref);
			u_r = pid_controller_r(vel_r, vel_r_ref);
			//u_l += pid_controller_l(vel_l, vel_l_ref);
			//u_r += pid_controller_r(vel_r, vel_r_ref);

			if(u_l>127) u_l=127;
			if(u_l<-127) u_l=-127;
			if(u_r>127) u_r=127;
			if(u_r<-127) u_r=-127;

			//printf("u_l=%5.1f, u_r=%5.1f   in vel_ctrl\n",u_l, u_r);
			motor_drive((int)u_l, (int)u_r);
		}

		//usleep(19000); // 19msec →　スケジューラで20msec weight と等価
		usleep((T_VEL_CTRL -1) *1000 ); // 1ms はスケジューラの精度で遅れる

	}

	motor_drive(0,0);

	pthread_exit(0);


}

static float pid_controller_l(float vel_cur, float vel_ref){

	float u;
	float err, err_dif;

	err =  vel_ref - vel_cur;
	//err = vel_cur - vel_ref;
	err_sum_l += err;
	err_dif = err - err_prev_l;
	err_prev_l = err;

	u = Kp_l * err + Ki_l * err_sum_l + Kd_l * err_dif;

	return u;
}

static float pid_controller_r(float vel_cur, float vel_ref){

	float u;
	float err, err_dif;

	err =  vel_ref - vel_cur;
	err_sum_r += err;
	err_dif = err - err_prev_r;
	err_prev_r = err;

	u = Kp_r * err + Ki_r * err_sum_r + Kd_r * err_dif;

	return u;
}

static int thread_create()
{
	pthread_attr_t attr;
	struct sched_param sched_param;
	int ret;
	int policy;

	pthread_attr_init(&attr);

	 // 優先度20、ポリシーFIFOでスレッド生成
	sched_param.sched_priority = 20;
	pthread_attr_setschedparam(&attr, &sched_param);
	pthread_setschedparam(pthread_self(), SCHED_FIFO , &sched_param);

	ret = pthread_create(&tasks[0], &attr, thread_code, (void *)2);

	pthread_getschedparam(pthread_self(), &policy, &sched_param);
	printf("plicy = %s, priority =%d\n",
		(policy == SCHED_FIFO)  ? "SCHED_FIFO" :
		(policy == SCHED_RR)    ? "SCHED_RR" :
		(policy == SCHED_OTHER) ? "SCHED_OTHER" :
		"???",
 		sched_param.sched_priority);

	return 0;
}

void vel_ctrl_set(float l_ref, float r_ref){

	vel_l_ref= l_ref;
	vel_r_ref= r_ref;
	vel_l_ref_org= l_ref;
	vel_r_ref_org= r_ref;
	//printf("vel_l_ref=%6.1f   vel_r_ref=%6.1f    in vel_ctrl_set \n",vel_l_ref, vel_r_ref);

	return;

}

void vel_ctrl_get_speed(float *speed_l, float *speed_r){

	*speed_l = vel_l;
	*speed_r = vel_r;
}


void vel_ctrl_mod(float l_mod, float r_mod){

	vel_l_ref = vel_l_ref_org + l_mod;
	vel_r_ref = vel_r_ref_org + r_mod;

	//printf("vel_l_ref=%6.1f, vel_r_ref=%6.1f   in vle_ctrl_mod\n",vel_l_ref, vel_r_ref);

	return;
}

void vel_ctrl_reset(){

	err_sum_l=0;
	err_sum_r=0;
	err_prev_l=0;
	err_prev_r=0;

	vel_l_ref=0;
	vel_r_ref=0;
	vel_l_ref_org=0;
	vel_r_ref_org=0;

	return;

}

void vel_ctrl_create(){

	//vel_ctrl_reset();
	thread_create();
	printf("Velocity Control thread created!     thread_id=%d\n",(int)tasks[0]);
	//printf("Velocity COntrol thread stopped\n");

}

void vel_ctrl_cancel(){

	motor_drive(0, 0);
	pthread_cancel(tasks[0]);

}


